/**
  *  \file timdist.cc
  *  \brief Compute distances, HOST.EXE way
  *
  *  This one uses a compiled-in version of the table. Changing this
  *  to use 'disttabl.dat' would be trivial: simply open the file
  *  and read it into 'square_table'.
  *
  *  To use this in your fuel computation, you would do
  *     way = ship.speed * ship.speed;
  *     if (ship.is_gravitonic)
  *        way *= 2:
  *     move = tim_dist(ship.waypoint_dx, ship.waypoint_dy);
  *     if (move > way)
  *        move = way;
  *     fuel_usage = engine[ship.eng_type].fuel_usage[ship.speed] * (ship.mass / 10) * move)
  *           / (way * 10000L);
  *
  *  This code is public domain.
  */

#include <cmath>
#include <cstdio>

const unsigned char bit_table[] = {
#include "bittable.inc"
};

const unsigned char square_table[168][21] = {
#include "disttabl.inc"
};

/** Compute distances, HOST.EXE way. This computes the distance for which
    fuel usage is computed for a waypoint of dx,dy ly. Since, if the distance
    exceeds the maximum distance allowed by the ship speed, only that maximum
    amount is used, this function yields an "imprecise" value if the
    dx,dy is outside that range. */
int tim_dist(int dx, int dy)
{
    /* normalize. The formula is symmetric, so we only handle positive
       cases. */
    if (dx < 0)
        dx = -dx;
    if (dy < 0)
        dy = -dy;

    /* we cannot handle dy==0, so swap dx/dy in this case. The formula
       is symmetric. If dx==dy==0, we know the result to be zero. */
    if (dy == 0) {
        if (dx == 0)
            return 0;
        dy = dx;
        dx = 0;
    }

    /* outside table? return 'imprecise' result in this case, where
       'imprecise' means it does not match HOST's behaviour. */
    if (dx >= 168 || dy >= 168)
        return int(std::sqrt(dx*dx + dy*dy));

    /* now, start at dy-1, and count bits in the table. Each set bit
       means the value in this cell is one larger than in the cell
       left to it. */
    int value = dy - 1;
    int index = 0;
    while (dx >= 8) {
        value += bit_table[square_table[dy-1][index++]];
        dx -= 8;
    }
    for (int bit = 0; bit <= dx; ++bit)
        if (square_table[dy-1][index] & (1 << bit))
            ++value;
    return value;
}

int main()
{
    for (int y = 0; y < 168; ++y) {
        for (int x = 0; x < 168; ++x) {
            std::printf("%3d %3d -> %4d\n", x, y, tim_dist(x, y));
        }
        std::puts("");
    }
}
